/*
 *  fs/ldiskfs/extents_status.h
 *
 * Written by Yongqiang Yang <xiaoqiangnk@gmail.com>
 * Modified by
 *	Allison Henderson <achender@linux.vnet.ibm.com>
 *	Zheng Liu <wenqing.lz@taobao.com>
 *
 */

#ifndef _LDISKFS_EXTENTS_STATUS_H
#define _LDISKFS_EXTENTS_STATUS_H

/*
 * Turn on ES_DEBUG__ to get lots of info about extent status operations.
 */
#ifdef ES_DEBUG__
#define es_debug(fmt, ...)	printk(fmt, ##__VA_ARGS__)
#else
#define es_debug(fmt, ...)	no_printk(fmt, ##__VA_ARGS__)
#endif

/*
 * With ES_AGGRESSIVE_TEST defined, the result of es caching will be
 * checked with old map_block's result.
 */
#define ES_AGGRESSIVE_TEST__

/*
 * These flags live in the high bits of extent_status.es_pblk
 */
#define ES_SHIFT	60

#define EXTENT_STATUS_WRITTEN	(1 << 3)
#define EXTENT_STATUS_UNWRITTEN (1 << 2)
#define EXTENT_STATUS_DELAYED	(1 << 1)
#define EXTENT_STATUS_HOLE	(1 << 0)

#define EXTENT_STATUS_FLAGS	(EXTENT_STATUS_WRITTEN | \
				 EXTENT_STATUS_UNWRITTEN | \
				 EXTENT_STATUS_DELAYED | \
				 EXTENT_STATUS_HOLE)

#define ES_WRITTEN		(1ULL << 63)
#define ES_UNWRITTEN		(1ULL << 62)
#define ES_DELAYED		(1ULL << 61)
#define ES_HOLE			(1ULL << 60)

#define ES_MASK			(ES_WRITTEN | ES_UNWRITTEN | \
				 ES_DELAYED | ES_HOLE)

struct ldiskfs_sb_info;
struct ldiskfs_extent;

struct extent_status {
	struct rb_node rb_node;
	ldiskfs_lblk_t es_lblk;	/* first logical block extent covers */
	ldiskfs_lblk_t es_len;	/* length of extent in block */
	ldiskfs_fsblk_t es_pblk;	/* first physical block */
};

struct ldiskfs_es_tree {
	struct rb_root root;
	struct extent_status *cache_es;	/* recently accessed extent */
};

extern int __init ldiskfs_init_es(void);
extern void ldiskfs_exit_es(void);
extern void ldiskfs_es_init_tree(struct ldiskfs_es_tree *tree);

extern int ldiskfs_es_insert_extent(struct inode *inode, ldiskfs_lblk_t lblk,
				 ldiskfs_lblk_t len, ldiskfs_fsblk_t pblk,
				 unsigned int status);
extern void ldiskfs_es_cache_extent(struct inode *inode, ldiskfs_lblk_t lblk,
				 ldiskfs_lblk_t len, ldiskfs_fsblk_t pblk,
				 unsigned int status);
extern int ldiskfs_es_remove_extent(struct inode *inode, ldiskfs_lblk_t lblk,
				 ldiskfs_lblk_t len);
extern void ldiskfs_es_find_delayed_extent_range(struct inode *inode,
					ldiskfs_lblk_t lblk, ldiskfs_lblk_t end,
					struct extent_status *es);
extern int ldiskfs_es_lookup_extent(struct inode *inode, ldiskfs_lblk_t lblk,
				 struct extent_status *es);

static inline int ldiskfs_es_is_written(struct extent_status *es)
{
	return (es->es_pblk & ES_WRITTEN) != 0;
}

static inline int ldiskfs_es_is_unwritten(struct extent_status *es)
{
	return (es->es_pblk & ES_UNWRITTEN) != 0;
}

static inline int ldiskfs_es_is_delayed(struct extent_status *es)
{
	return (es->es_pblk & ES_DELAYED) != 0;
}

static inline int ldiskfs_es_is_hole(struct extent_status *es)
{
	return (es->es_pblk & ES_HOLE) != 0;
}

static inline unsigned int ldiskfs_es_status(struct extent_status *es)
{
	return es->es_pblk >> ES_SHIFT;
}

static inline ldiskfs_fsblk_t ldiskfs_es_pblock(struct extent_status *es)
{
	return es->es_pblk & ~ES_MASK;
}

static inline void ldiskfs_es_store_pblock(struct extent_status *es,
					ldiskfs_fsblk_t pb)
{
	ldiskfs_fsblk_t block;

	block = (pb & ~ES_MASK) | (es->es_pblk & ES_MASK);
	es->es_pblk = block;
}

static inline void ldiskfs_es_store_status(struct extent_status *es,
					unsigned int status)
{
	es->es_pblk = (((ldiskfs_fsblk_t)
			(status & EXTENT_STATUS_FLAGS) << ES_SHIFT) |
		       (es->es_pblk & ~ES_MASK));
}

static inline void ldiskfs_es_store_pblock_status(struct extent_status *es,
					       ldiskfs_fsblk_t pb,
					       unsigned int status)
{
	es->es_pblk = (((ldiskfs_fsblk_t)
			(status & EXTENT_STATUS_FLAGS) << ES_SHIFT) |
		       (pb & ~ES_MASK));
}

extern void ldiskfs_es_register_shrinker(struct ldiskfs_sb_info *sbi);
extern void ldiskfs_es_unregister_shrinker(struct ldiskfs_sb_info *sbi);
extern void ldiskfs_es_lru_add(struct inode *inode);
extern void ldiskfs_es_lru_del(struct inode *inode);

#endif /* _LDISKFS_EXTENTS_STATUS_H */
